# -*- coding: utf-8 -*-
import requests
import re
import time
import getpass # To securely get password input
from atproto import Client, models # Make sure this is installed: pip3 install atproto requests

def parse_usernames_from_file(filepath):
    """Reads the Following.txt file and extracts usernames."""
    usernames = set() # Use a set to avoid duplicates
    username_pattern = re.compile(r"^Username:\s*(.+)$")
    print(f"Reading usernames from {filepath}...")
    try:
        # Try reading with UTF-8 first
        with open(filepath, 'r', encoding='utf-8') as f:
            content = f.readlines()
    except UnicodeDecodeError:
        try:
            # Fallback to latin-1 if UTF-8 fails
            print("UTF-8 failed, trying latin-1 encoding...")
            with open(filepath, 'r', encoding='latin-1') as f:
                 content = f.readlines()
        except FileNotFoundError:
            print(f"Error: File not found at {filepath}")
            return []
        except Exception as e:
            print(f"Error reading file {filepath} with latin-1: {e}")
            return []
    except FileNotFoundError:
        print(f"Error: File not found at {filepath}")
        return []
    except Exception as e:
        print(f"Error reading file {filepath}: {e}")
        return []

    for line in content:
         match = username_pattern.match(line)
         if match:
             username = match.group(1).strip()
             # Handle usernames ending with ... more robustly
             while username.endswith('.'):
                  username = username[:-1]
             # Exclude 'N/A' and ensure username is not empty
             if username and username.lower() != 'n/a':
                 # Remove potential leading/trailing quotes if present
                 username = username.strip('\'"')
                 usernames.add(username)

    print(f"Extracted {len(usernames)} unique usernames.")
    return list(usernames)

def search_and_follow_bluesky_manual(client: Client, usernames: list):
    """Searches Bluesky for usernames and prompts user to follow ALL matches."""
    confirmed_followed_count = 0
    skipped_count = 0
    error_count = 0
    already_following_count = 0
    not_found_count = 0
    rate_limit_hit = False
    processed_count = 0 # Initialize processed_count

    total_users = len(usernames)
    print(f"\nSearching Bluesky and prompting for follows (manual confirmation required)...")

    for i, username in enumerate(usernames, 1):
        processed_count = i # Keep track of the last processed user index
        print(f"\n--- Processing TikTok user: {username} ({i}/{total_users}) ---")
        try:
            # Use the authenticated client's search method
            profile_search = client.app.bsky.actor.search_actors(params={'q': username, 'limit': 10})
            time.sleep(0.1) # Small delay after search API call

            if profile_search and profile_search.actors:
                actors = profile_search.actors
                actor_count = len(actors)
                display_name = lambda actor: f"'{actor.display_name}'" if actor.display_name else '(No display name)'

                # --- Always prompt if one or more matches are found ---
                print(f"Found {actor_count} potential match(es):")
                # Simple sorting/filtering (optional, could be refined)
                potential_handles = [username.lower(), f"{username.lower()}.bsky.social"]
                exact_matches = [a for a in actors if a.handle.lower() in potential_handles]
                other_matches = [a for a in actors if a not in exact_matches]
                sorted_actors = exact_matches + other_matches

                followed_this_user_session = False # Track if we followed someone for this TikTok user in this session
                processed_actors_count = 0
                # Use a variable to capture the last response in the inner loop
                follow_confirm = ''

                for idx, actor in enumerate(sorted_actors):
                    processed_actors_count += 1
                    print(f"  {idx+1}. Handle: @{actor.handle}, Name: {display_name(actor)}")

                    # Check if already following
                    is_following = False
                    try:
                        relationship = client.app.bsky.graph.get_relationships(
                            params={'actor': client.me.did, 'others': [actor.did]}
                        )
                        time.sleep(0.1)
                        if relationship and relationship.relationships and len(relationship.relationships) > 0:
                           rel_data = relationship.relationships[0]
                           if hasattr(rel_data, 'following') and rel_data.following:
                               is_following = True
                        if is_following:
                            print("     (Already following)")
                            # If it's the only match and we're already following, count it here
                            if actor_count == 1:
                                 already_following_count += 1
                            continue # Check next potential match for this TikTok user
                    except Exception as rel_err:
                         if 'ratelimiterror' in str(rel_err).lower() or 'rate limit' in str(rel_err).lower():
                             print(f"\nRate limit hit during relationship check for {username}. Stopping.")
                             rate_limit_hit = True
                             break # Break inner actor loop
                         else:
                             print(f"     (Could not check follow status: {type(rel_err).__name__})")

                    # Prompt user (Always happens if not already following)
                    while True:
                        follow_confirm = input(f"  Follow @{actor.handle}? (y/n/s=skip ALL for {username}): ").lower()
                        if follow_confirm == 'y':
                            try:
                                print(f"  Attempting to follow @{actor.handle}...")
                                # *** THE FIX IS HERE ***
                                response = client.follow(actor.did) # Pass DID directly
                                print(f"  Successfully followed @{actor.handle}")
                                confirmed_followed_count += 1
                                followed_this_user_session = True
                                time.sleep(2.5) # Delay
                                break # Exit prompt loop for this actor
                            except Exception as follow_err:
                                if 'ratelimiterror' in str(follow_err).lower() or 'rate limit' in str(follow_err).lower():
                                     print(f"\nRate limit hit during follow attempt for {username}. Stopping.")
                                     rate_limit_hit = True
                                else:
                                     print(f"  Error following @{actor.handle}: {type(follow_err).__name__} - {follow_err}")
                                     error_count += 1
                                # Stop trying for this TikTok user if follow fails
                                followed_this_user_session = True # Treat as processed to prevent skip count increment later
                                break # Exit prompt loop for this actor
                        elif follow_confirm == 'n':
                            print(f"  Skipping @{actor.handle}.")
                            break # Exit prompt loop for this actor, check next match
                        elif follow_confirm == 's':
                            print(f"  Skipping all remaining checks for {username}.")
                            skipped_count += (actor_count - processed_actors_count) # Add remaining actors for this user to skip count
                            followed_this_user_session = True # Mark as processed to break outer loop
                            break # Exit prompt loop for this actor
                        else:
                            print("  Invalid input. Please enter 'y', 'n', or 's'.")

                    if rate_limit_hit or follow_confirm == 's' or (follow_confirm == 'y' and followed_this_user_session):
                       # If we followed someone OR skipped all remaining, move to the next TikTok username
                       break # Exit the loop iterating through actors for *this TikTok username*

                if rate_limit_hit: break # Break outer username loop if rate limit hit in inner loop

                # Increment skip count logic (minor adjustment for clarity)
                if not followed_this_user_session and follow_confirm != 's':
                    # Check if the loop finished without following, and the last action wasn't skipping all
                    # Also ensure we didn't just skip because they were already followed
                    is_already_following = False
                    if 'is_following' in locals(): # Check if variable exists from loop
                        is_already_following = is_following
                    if not is_already_following:
                        skipped_count += 1


            else:
                print(f"No profiles found matching '{username}'.")
                not_found_count += 1

        except Exception as search_err:
            if 'ratelimiterror' in str(search_err).lower() or 'rate limit' in str(search_err).lower():
                 print(f"\nRate limit hit during search for {username}. Stopping.")
                 rate_limit_hit = True
                 break # Stop the whole process
            else:
                 print(f"Error searching/processing {username}: {type(search_err).__name__} - {search_err}")
                 error_count += 1
                 # Optional: break on other persistent errors if needed

    # --- Summary ---
    # Use the tracked processed_count for accuracy, especially if the loop broke early
    print("\n--- Search and Follow Summary (Manual Confirmation Only) ---")
    print(f"Processed {processed_count} out of {total_users} TikTok usernames.")
    print(f"Followed after prompt ('y'): {confirmed_followed_count}")
    print(f"Already following: {already_following_count}")
    print(f"Not Found on Bluesky: {not_found_count}")
    # Combine skips and errors for simplicity
    print(f"Skipped (manual 'n'/'s' or partial checks) or Errored: {skipped_count + error_count}")
    print(f"Encountered errors during search/follow attempts: {error_count}")
    if rate_limit_hit:
         print("Processing stopped early due to hitting API rate limits.")
    elif processed_count < total_users and total_users > 0: # Check if loop broke early for other reasons
         print(f"Processing stopped early after {processed_count} users.")

# --- Main Execution ---
file_path = 'Following.txt' # Make sure this points to your file

print("--- Bluesky Follower Tool (v4.1: Fixed follow method call) ---")
# Optional: Suppress the SSL warning
# import warnings
# from urllib3.exceptions import NotOpenSSLWarning
# warnings.filterwarnings('ignore', category=NotOpenSSLWarning)
# print("Note: Suppressed NotOpenSSLWarning from urllib3.")

print("⚠️ Use an App Password from Bluesky Settings.")
print("⚠️ Manual confirmation ('y') required for ALL follows.")
print("⚠️ Excessive use may lead to rate limits or account flags.")

bluesky_handle = input("Enter your Bluesky handle (e.g., yourname.bsky.social): ")
bluesky_password = getpass.getpass("Enter your Bluesky App Password: ")

client = Client()
try:
    print("\nLogging into Bluesky...")
    profile = client.login(bluesky_handle, bluesky_password)
    print(f"Successfully logged in as @{profile.handle}")
except Exception as login_err:
    # --- THIS BLOCK HANDLES LOGIN ERRORS ---
    # It IS indented under the 'except' line
    if 'ModelError' in str(login_err) and 'emailConfirmed' in str(login_err):
         print(f"\nLogin Failed: {login_err}")
         print("This might indicate an outdated 'atproto' library. Try updating:")
         print("python3 -m pip install --upgrade atproto")
    else:
         print(f"\nLogin Failed: {type(login_err).__name__} - {login_err}")
         print("Please check your handle and App Password.")
    print("Exiting.")
    exit()
    # --- END OF LOGIN ERROR BLOCK ---

# --- CODE CONTINUES AFTER LOGIN ATTEMPT ---
# These lines are NOT indented, so they run after the try/except block finishes
tiktok_usernames = parse_usernames_from_file(file_path)

if tiktok_usernames and client.me:
    # This block IS indented under the 'if'
    # Optional: Limit the number of searches for testing
    # tiktok_usernames = tiktok_usernames[:25]

    # Call the correct function name for the manual version
    search_and_follow_bluesky_manual(client, tiktok_usernames)
else:
    # This block IS indented under the 'else'
    if not tiktok_usernames:
        print("No valid usernames were extracted from the file.")
    # Added a check here - client.me might not exist if login failed AND the except block was somehow bypassed (unlikely)
    elif not hasattr(client, 'me') or not client.me:
         print("Client login failed or user profile not available, cannot proceed.")

# This line is NOT indented
print("\nScript finished.")

